def all(a, axis=None, out=None, keepdims=False):
    # TODO(okuta): check type
    return a.all(axis=axis, out=out, keepdims=keepdims)


def any(a, axis=None, out=None, keepdims=False):
    # TODO(okuta): check type
    return a.any(axis=axis, out=out, keepdims=keepdims)
